/**
 * @file 	LMPApproximator.java
 * 
 * @date 	Jan 15, 2005
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.lmp.approx;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import ca.mcgill.lmp.MeasurableSet;
import ca.mcgill.lmp.LMP;
import ca.mcgill.lmp.MeasurableFunction;
import ca.mcgill.lmp.Measure;
import ca.mcgill.lmp.State;
import ca.mcgill.lmp.util.DiscreteSet;
import ca.mcgill.lmp.util.MultiInterval;
import ca.mcgill.lmp.util.kernels.DiscreteKernel;

/**
 *
 */
public class LMPApproximator
{
    /**
     * Comment for <code>lmp</code>
     */
    private LMP lmp;
    
    /**
     * Comment for <code>monteCarlo</code>
     */
    private MonteCarlo monteCarlo;
    
    /**
     * @param lmp
     * @param underlyingMeasure
     */
    public LMPApproximator(LMP lmp, Measure underlyingMeasure)
    {
        this.lmp = lmp;
        monteCarlo = new MonteCarlo(underlyingMeasure);
    }
    
    /**
     * @param epsilon
     * @param n
     * @return
     */
    public LMP approximate(double epsilon, int n)
    {
        DiscreteSet approxStateSpace = new DiscreteSet();
        Collection previousLevel = new ArrayList();
        DiscreteKernel [] approximationKernels = new DiscreteKernel[lmp.getNumberOfLabels()];
        for (int it = 0; it < approximationKernels.length; it++)
        {
            approximationKernels[it] = new DiscreteKernel();
        }
        // Construct level 0
        State startState = new ApproximationState(lmp.getStateSpace(), 0);
        previousLevel.add(startState);
        approxStateSpace.add(startState);
        // Construct iteratively the other levels
        for (int l = 1; l < n; l++)
        {
            // FIRST STEP: compute the set of state for this new level
            Collection currentLevel = new ArrayList();
            List generatingPartitions = new ArrayList();
            Collection partition = generateUnitIntervalPartition(epsilon, previousLevel.size());
            
            // for all actions
            for (int a = 0; a < lmp.getNumberOfLabels(); a++)
            { 
                List currentPartition = new ArrayList();
                generatingPartitions.add(currentPartition);
                
	            // for every states in the previous level
	            Iterator previousLevelStatesIterator = previousLevel.iterator();
	            while (previousLevelStatesIterator.hasNext())
	            {
	                MeasurableSet C = ((ApproximationState) previousLevelStatesIterator.next()).getSet();
	                // Iterator thru the intervals {0}, (0, epsilon/m], (epsilon/m, 2 epsilon/m], ... (where m = size of prev level)
	                Iterator unitIntervalPartitionIterator = partition.iterator();
	                while (unitIntervalPartitionIterator.hasNext())
	                {
	                    MultiInterval currentInterval = (MultiInterval) unitIntervalPartitionIterator.next();

	                    MeasurableSet set =  new InverseSet(lmp.getTransitionKernel(a).getTransitionFunction(C), currentInterval);
//	                    boolean toAdd = true;
//	                    // check that this generator is not null
//	                    if (monteCarlo.isEmpty(set))
//	                    {
//	                        toAdd = false;
//	                    }
//	                    // check that this generator is not already present
//                        for (int in = 0; in < generatingPartitions.size() && toAdd; in++)
//                        {
//                            List currentPartition2 = (List) generatingPartitions.get(in);
//                            for (int in2 = 0; in2 < currentPartition2.size(); in2++)
//                            {
//                                MeasurableSet currentOldSet = (MeasurableSet) currentPartition2.get(in2);
//	                            if (monteCarlo.equals(currentOldSet, set))
//	                            {
//	                                toAdd = false;
//	                            }
//                            }
//                        }
//                        if (toAdd)
//                        {
                        currentPartition.add(set);
//                        }
                    }
                }
            }
            // construct the partition generated by the sets in generatingSets
            Iterator membershipKeysIterator = generateMembershipKeys(generatingPartitions).iterator();
            while (membershipKeysIterator.hasNext())
            {
                MembershipKeys currentKey = (MembershipKeys) membershipKeysIterator.next();
                GeneratedSet currentSet = new GeneratedSet(generatingPartitions, currentKey);
                // create a new state for each non-empty
                if (!monteCarlo.isEmpty(currentSet))
                {
                    currentLevel.add(new ApproximationState(currentSet, l));
                }
            }
            // SECOND STEP: compute the transition probabilities b/w the new level and the prev. one
            Iterator currentLevelStatesIterator = currentLevel.iterator();
            while (currentLevelStatesIterator.hasNext())
            {
                ApproximationState sourceState = (ApproximationState) currentLevelStatesIterator.next();
                MeasurableSet B = sourceState.getSet();
                Iterator previousLevelStatesIterator = previousLevel.iterator();
                while (previousLevelStatesIterator.hasNext())
                {
                    ApproximationState destinationState = (ApproximationState) previousLevelStatesIterator.next();
                    for (int a = 0; a < lmp.getNumberOfLabels(); a++)
                    {              
	                    MeasurableSet X = destinationState.getSet();
	                    double currentProbability = monteCarlo.inf(lmp.getTransitionKernel(a).getTransitionFunction(X), B);
	                    approximationKernels[a].setProbability(sourceState, destinationState, currentProbability);
                    }
                }
            }
            approxStateSpace.addAll(currentLevel);
            previousLevel = currentLevel;
        }
        return new LMP(approxStateSpace, approximationKernels, startState);
    }
    
    /**
     * @param epsilon
     * @param m
     * @return
     */
    private Collection generateUnitIntervalPartition(double epsilon, int m)
    {
        double increment = epsilon / (double) m;
        List partition = new ArrayList();
        partition.add(new MultiInterval(0, 0));
        double currentLowerBound = 0;
        double currentUpperBound = increment;
        while (currentUpperBound < 1)
        {
            partition.add(new MultiInterval(currentLowerBound, currentUpperBound, false, true));
            currentLowerBound = currentUpperBound;
            currentUpperBound = currentUpperBound + increment;
        }
        partition.add(new MultiInterval(currentLowerBound, 1, false, true));
        return partition;
    }
    
    /**
     * @param n
     * @return
     */
    private Collection generateMembershipKeys(List generatingPartitions)
    {
        int [] lengths = new int[generatingPartitions.size()];
        int [] prefix = new int[generatingPartitions.size()];
        for (int i = 0; i < generatingPartitions.size(); i++)
        {
            lengths[i] = ((List) generatingPartitions.get(i)).size();
            prefix[i] = 0;
        }
        
        return recGenerateMembershipKeys(lengths, prefix, 0);
    }
    
    private Collection recGenerateMembershipKeys(int [] lengths, int [] prefix, int index)
    {
        Collection returnedValue = new ArrayList();
        if (index == lengths.length)
        {
            returnedValue.add(new MembershipKeys(prefix));
            return returnedValue;
        }
        for (int i = 0; i < lengths[index]; i++)
        {
            int [] newPrefix = new int[prefix.length];
            for (int j = 0; j < prefix.length; j++)
            {
                newPrefix[j] = prefix[j];
            }
            newPrefix[index] = i;
            returnedValue.addAll(recGenerateMembershipKeys(lengths, newPrefix, index + 1));
        }
        return returnedValue;
    }
    
    /**
     * 
     *
     */
    class MembershipKeys
    {
        /**
         * Comment for <code>keys</code>
         */
        private int [] keys;
        
        /**
         * @param keys
         */
        public MembershipKeys(int [] keys)
        {
            this.keys = keys;
        }
        
        /**
         * @param n
         * @return
         */
        public int get(int n)
        {
            return keys[n];
        }
        
        public String toString()
        {
            StringBuffer buffer = new StringBuffer();
            buffer.append("[");
            for (int i = 0; i < keys.length; i++)
            {
                buffer.append("" + keys[i] + " ");
            }
            buffer.append("]");
            return buffer.toString();
        }
    }
    
    /**
     * 
     *
     */
    class InverseSet implements MeasurableSet
    {
        /**
         * Comment for <code>f</code>
         */
        private MeasurableFunction f;
        
        /**
         * Comment for <code>interval</code>
         */
        private MultiInterval interval;
        
        /**
         * @param f
         * @param interval
         */
        public InverseSet(MeasurableFunction f, MultiInterval interval)
        {
            this.interval = interval;
            this.f = f;
        }
        
        /**
         * @param s
         * @return
         */
        public boolean contains(State s)
        {
            // note: if f.evaluateAt(s) is close to the boundaries of the interval, more care most be taken...
            if (interval.contains(f.evaluateAt(s)))
            {
                return true;
            }
            return false;
        }
        
        public String toString()
        {
            return "(" + f.toString() + ")^-1 (" + interval.toString() + ")";
        }
    }
    
    /**
     * 
     *
     */
    class GeneratedSet implements MeasurableSet
    {
        /**
         * Comment for <code>generators</code>
         */
        private List generatingPartitions;
        
        /**
         * Comment for <code>membershipKeys</code>
         */
        private MembershipKeys membershipKeys;
        
        /**
         * @param generators
         * @param membershipKeys
         */
        public GeneratedSet(List generatingPartitions, MembershipKeys membershipKeys)
        {
            this.generatingPartitions = generatingPartitions;
            this.membershipKeys = membershipKeys;
        }
        
        /**
         * @param s
         * @return
         */
        public boolean contains(State s)
        {
            for (int i = 0; i < generatingPartitions.size(); i++)
            {
                List currentPartition = (List) generatingPartitions.get(i);
                MeasurableSet currentSet = (MeasurableSet) currentPartition.get(membershipKeys.get(i));
                if (!currentSet.contains(s))
                {
                    return false;
                }
            }
            return true;
        }
        
        public String toString()
        {
            StringBuffer buffer = new StringBuffer();
            for (int i = 0; i < generatingPartitions.size(); i++)
            {
                List currentPartition = (List) generatingPartitions.get(i);
                Object currentSet = currentPartition.get(membershipKeys.get(i));
                buffer.append(currentSet.toString());
                if (i != generatingPartitions.size() - 1)
                {
                    buffer.append(" INTER ");
                }
            }
            return buffer.toString();
        }
    }
}