/**
 * @file	MonteCarlo.java
 * 
 * @date	Feb 3, 2005
 * @author	Alexandre Bouchard
 */
package ca.mcgill.lmp.approx;

import ca.mcgill.lmp.EmptySetException;
import ca.mcgill.lmp.MeasurableFunction;
import ca.mcgill.lmp.MeasurableSet;
import ca.mcgill.lmp.Measure;
import ca.mcgill.lmp.State;



/**
 * 
 *
 */
public class MonteCarlo
{
    /**
     * Comment for <code>monteCarloTrials</code>
     */
    private int monteCarloTrials = 50;
    public static int DEFAULT_NUMBER_OF_TRIALS = 50;
    
    /**
     * The space (set on which the sampling measure is defined)
     */
    private MeasurableSet X;
    
    /**
     * Comment for <code>samplingMeasure</code>
     */
    private Measure samplingMeasure;
    
    /**
     * @param X
     */
    public MonteCarlo(Measure samplingMeasure)
    {
        this.samplingMeasure = samplingMeasure;
        this.X = samplingMeasure.getX();
    }
    
    /**
     * @param f
     * @param D
     * @return
     */
    public double integrate(MeasurableFunction f, MeasurableSet D)
    {
        double partialSum = 0;
        for (int counter = 0; counter < monteCarloTrials; counter++)
        {
            try
            {
                State currentState = samplingMeasure.sampleIn(X);
	            if (D.contains(currentState))
	            {
	                partialSum = partialSum + f.evaluateAt(currentState);
	            }   
            }
            catch (EmptySetException ese)
            {
                return 0;
            }            
        }
        return partialSum / (double) monteCarloTrials;
    }
    
    public double integrate(MeasurableFunction f)
    {
        return integrate(f, X);
    }
    
    /**
     * Returns an approximation the essential infimum of a function f on a set D.
     * We take the convention that if the given set is empty, we return 0
     * 
     * @param f
     * @param D
     * @return
     */
    public double inf(MeasurableFunction f, MeasurableSet D)
    {
        double partialMin = Double.MAX_VALUE;
        for (int counter = 0; counter < monteCarloTrials; counter++)
        {
            try
            {            
	            double currentValue = f.evaluateAt(samplingMeasure.sampleIn(D));
	            if (currentValue < partialMin)
	            {
	                partialMin = currentValue;
	            }
            }
            catch (EmptySetException ese)
            {
                return 0;
            }  	            
        }
        return partialMin;
    }
    
    /**
     * @param f
     * @return
     */
    public double inf(MeasurableFunction f)
    {
        return inf(f, X);
    }
    
    /**
     * @param D
     * @return
     */
    public boolean isEmpty(MeasurableSet D)
    {
        for (int counter = 0; counter < monteCarloTrials; counter++)
        {
            if (D.contains(samplingMeasure.sample()))
            {
                return false;
            }
        }
        return true;
    }
    
    public boolean equals(MeasurableSet D1, MeasurableSet D2)
    {
        for (int counter = 0; counter < monteCarloTrials; counter++)
        {
            State currentState = samplingMeasure.sample();
            if (D1.contains(currentState) && !D2.contains(currentState))
            {
                return false;
            }
            if (!D1.contains(currentState) && D2.contains(currentState))
            {
                return false;
            }
        }
        return true;
    }
    
    /**
     * @return Returns the monteCarloTrials.
     */
    public int getMonteCarloTrials()
    {
        return monteCarloTrials;
    }
    
    /**
     * @param monteCarloTrials The monteCarloTrials to set.
     */
    public void setMonteCarloTrials(int monteCarloTrials)
    {
        this.monteCarloTrials = monteCarloTrials;
    }
    
    public String toString()
    {
        return "Sampling measure: " + samplingMeasure.toString() + "\n"
        	+ "Number of trials: " + monteCarloTrials;
    }
}
